use rand::Rng;


pub fn do_challenge(stat: u8) -> (bool, u8) {
    if stat <= 20 {
        let mut rng = rand::thread_rng();
        let dice_roll = rng.gen_range(1..21);
        return match dice_roll {
            20 => { (false, 0) }
            1 => { (true, dice_roll) }
            _ => {
                if dice_roll <= stat {
                    (true, stat - dice_roll)
                } else {
                    (false, 0)
                }
            }
        };
    }
    (false, 0)
}

#[allow(dead_code)]
pub struct PlayerStats {
    // dungeon slayer attributes
    pub(crate) body: u8,
    pub(crate) agility: u8,
    pub(crate) mind: u8,

    // dungeon slayer properties
    //  for body
    pub(crate) strength: u8,
    pub(crate) toughness: u8,
    //  for agility
    pub(crate) movement: u8,
    pub(crate) dexterity: u8,
    // for mind
    pub(crate) wisdom: u8,
    // TODO maybe the wrong translation of Verstand
    pub(crate) aura: u8,
}

impl PlayerStats {
    /// create random values for the 3 attributes of the dungeon slayer character
    /// Each attribute is between 4 and 8, not over 20 in sum.
    fn create_random_attributes() -> (u8, u8, u8) {
        let mut values: [u8; 3] = [8, 8, 8];
        let mut rng = rand::thread_rng();
        while values[0] + values[1] + values[2] > 20 {
            let i = rng.gen_range(0..3);
            values[i] = values[i].saturating_sub(1);
        }
        (values[0], values[1], values[2])
    }

    /// create random values for the 6 properties of the dungeon slayer character
    /// Each property is between 0 and 4, not over 8 in sum.
    fn create_random_properties() -> (u8, u8, u8, u8, u8, u8) {
        let mut values: [u8; 6] = [4, 4, 4, 4, 4, 4];
        let mut rng = rand::thread_rng();
        while values[0] + values[1] + values[2] + values[3] + values[4] + values[5] > 8 {
            let i = rng.gen_range(0..6);
            values[i] = values[i].saturating_sub(1);
        }
        (values[0], values[1], values[2], values[3], values[4], values[5])
    }

    /// create a set of player stats containing attributes and properties according to the
    /// dungeon slayer rules.
    pub fn create_random() -> PlayerStats {
        let attributes = PlayerStats::create_random_attributes();
        let properties = PlayerStats::create_random_properties();
        PlayerStats {
            body: attributes.0,
            agility: attributes.1,
            mind: attributes.2,
            strength: properties.0,
            toughness: properties.1,
            movement: properties.2,
            dexterity: properties.3,
            wisdom: properties.4,
            aura: properties.5,
        }
    }

    /// calculate the max life based on this player stats.
    pub fn get_max_life(&self) -> u8 {
        self.body + self.toughness + 10
    }
    pub fn get_defense(&self) -> u8 {
        self.body + self.toughness // TODO + self.armor
    }
    pub fn get_hit(&self) -> u8 {
        self.body + self.strength // TODO + self.weapon
    }
}

#[allow(dead_code)]
pub struct MonsterStats {
    // dungeon slayer attributes
    pub(crate) body: u8,
    pub(crate) agility: u8,
    pub(crate) mind: u8,

    // dungeon slayer properties
    //  for body
    pub(crate) strength: u8,
    pub(crate) toughness: u8,
    //  for agility
    pub(crate) movement: u8,
    pub(crate) dexterity: u8,
    // for mind
    pub(crate) wisdom: u8,
    // TODO maybe the wrong translation of Verstand
    pub(crate) aura: u8,

    pub(crate) max_life: u8,
    pub(crate) defense: u8,
    pub(crate) initiative: u8,
    pub(crate) walk: u8,
    pub(crate) hit: u8,
    pub(crate) shoot: u8,
    pub(crate) cast: u8,
    pub(crate) targeted_cast: u8,
}

#[test]
fn test_create_random_attributes() {
    for _ in 0..1000 {
        let attributes = PlayerStats::create_random_attributes();
        assert!(attributes.0 <= 8);
        assert!(attributes.1 <= 8);
        assert!(attributes.2 <= 8);
        assert_eq!(attributes.0 + attributes.1 + attributes.2, 20);
    }
}

#[test]
fn test_create_random_properties() {
    for _ in 0..1000 {
        let properties = PlayerStats::create_random_properties();
        assert!(properties.0 <= 4);
        assert!(properties.1 <= 4);
        assert!(properties.2 <= 4);
        assert!(properties.3 <= 4);
        assert!(properties.4 <= 4);
        assert!(properties.5 <= 4);
        assert_eq!(properties.0 + properties.1 + properties.2 + properties.3 + properties.4 + properties.5, 8);
    }
}