use ratatui::prelude::Color;

use crate::dungeon_slayer::MonsterStats;
use crate::level::Level;
use crate::position::{HasPosition, Position};
use rand::Rng;


pub trait Monster: HasPosition {
    fn get_name(&self) -> &str;
    fn is_dead(&self) -> bool;
    fn get_representation(&self) -> (&str, Color);
    fn decrease_life(&mut self, by: usize);
    // fn get_immutable_position(&self) -> &Position;
    fn get_experience_gain(&self) -> usize;
    fn get_ticks_between_steps(&self) -> u128;
    fn get_defense(&self) -> u8;
    fn get_hit(&self) -> u8;
    fn get_next_move(&self, level: &Level, player_pos: &Position) -> (i16, i16);
    #[cfg(test)]
    fn get_life(&self) -> usize;
}

macro_rules! default_monster {
    ($($t:ty),+ $(,)?) => ($(
        impl Monster for $t {
            fn get_name(&self) -> &str { &self.name }
            fn is_dead(&self) -> bool { self.life <= 0 }
            fn get_experience_gain(&self) -> usize { self.experience_gain }
            fn get_representation(&self) -> (&str, Color) { (&self.symbol, self.color) }
            fn decrease_life(&mut self, by: usize) {
                self.life = self.life.saturating_sub(by);
            }
            fn get_ticks_between_steps(&self) -> u128 { self.ticks_between_steps }
            fn get_defense(&self) -> u8 { self.monster_stats.defense }
            fn get_hit(&self) -> u8 { self.monster_stats.hit }
            fn get_next_move(&self, level: &Level, player_pos: &Position) -> (i16, i16) {
                self.calc_move(level, player_pos)
            }
            #[cfg(test)]
            fn get_life(&self) -> usize { self.life }
        }
        impl HasPosition for $t {
            fn get_position(&mut self) -> &mut Position {
                &mut self.position
            }
            fn get_immutable_position(&self) -> &Position {
                &self.position
            }
        }
    )+)
}

/// Lower Daemon
///     Page: 107
///     GH: 1
pub struct LowerDaemon {
    name: String,
    life: usize,
    position: Position,
    symbol: String,
    color: Color,
    experience_gain: usize,
    ticks_between_steps: u128,
    monster_stats: MonsterStats,
}

impl crate::monster::LowerDaemon {
    pub fn new_with_position(position: Position) -> Self {
        let monster_stats = MonsterStats {
            body: 5,
            agility: 5,
            mind: 5,
            strength: 2,
            toughness: 2,
            movement: 2,
            dexterity: 2,
            wisdom: 2,
            aura: 2,
            max_life: 9,
            defense: 9,
            initiative: 7,
            walk: 3,
            hit: 8,
            shoot: 0,
            cast: 0,
            targeted_cast: 0,
        };
        Self {
            name: "lower daemon".to_string(),
            life: monster_stats.max_life as usize,
            position,
            symbol: String::from("d"),
            color: Color::Black,
            experience_gain: 104,
            ticks_between_steps: 5,
            monster_stats,
        }
    }

    fn calc_move(&self, _level: &Level, player_pos: &Position) -> (i16, i16) {
        let d = self.position.distance(player_pos);
        if d == 1 {
            return self.position.get_direction(player_pos);
        }
        match rand::thread_rng().gen_range(0..5) {
            1 => { (1, 0) }
            2 => { (-1, 0) }
            3 => { (0, 1) }
            4 => { (0, -1) }
            _ => { (0, 0) }
        }
    }
}
default_monster!(LowerDaemon);

/// Rat
///     Page: 119
///     GH: 1
pub struct Rat {
    name: String,
    life: usize,
    position: Position,
    symbol: String,
    color: Color,
    experience_gain: usize,
    ticks_between_steps: u128,
    monster_stats: MonsterStats,
}

impl Rat {
    pub fn new_with_position(position: Position) -> Self {
        let monster_stats = MonsterStats {
            body: 2,
            agility: 4,
            mind: 1,
            strength: 1,
            toughness: 0,
            movement: 2,
            dexterity: 0,
            wisdom: 0,
            aura: 0,
            max_life: 3,
            defense: 2,
            initiative: 6,
            walk: 3,
            hit: 4,
            shoot: 0,
            cast: 0,
            targeted_cast: 0,
        };
        Self {
            name: "rat".to_string(),
            life: monster_stats.max_life as usize,
            position,
            symbol: String::from("R"),
            color: Color::Black,
            experience_gain: 26,
            ticks_between_steps: 5,
            monster_stats,
        }
    }
    fn calc_move(&self, _level: &Level, _player_pos: &Position) -> (i16, i16) {
        match rand::thread_rng().gen_range(0..5) {
            1 => { (1, 0) }
            2 => { (-1, 0) }
            3 => { (0, 1) }
            4 => { (0, -1) }
            _ => { (0, 0) }
        }
    }
}
default_monster!(Rat);

/// Orc
///     Page: 118
///     GH: 2
pub struct Orc {
    name: String,
    life: usize,
    position: Position,
    symbol: String,
    color: Color,
    experience_gain: usize,
    ticks_between_steps: u128,
    monster_stats: MonsterStats,
}

impl Orc {
    pub fn new_with_position(position: Position) -> Self {
        let monster_stats = MonsterStats {
            body: 10,
            agility: 6,
            mind: 2,
            strength: 2,
            toughness: 3,
            movement: 0,
            dexterity: 3,
            wisdom: 1,
            aura: 0,
            max_life: 23,
            defense: 14,
            initiative: 1,
            walk: 4,
            hit: 13,
            shoot: 10,
            cast: 0,
            targeted_cast: 0,
        };
        Self {
            name: "orc".to_string(),
            life: monster_stats.max_life as usize,
            position,
            symbol: String::from("O"),
            color: Color::DarkGray,
            experience_gain: 63,
            ticks_between_steps: 10,
            monster_stats,
        }
    }
    fn calc_move(&self, _level: &Level, _player_pos: &Position) -> (i16, i16) {
        match rand::thread_rng().gen_range(0..5) {
            1 => { (1, 0) }
            2 => { (-1, 0) }
            3 => { (0, 1) }
            4 => { (0, -1) }
            _ => { (0, 0) }
        }
    }
}

default_monster!(Orc);


#[test]
fn monsters_can_move() {
    let mut m = Rat::new_with_position(Position::new(0, 0, 0));
    assert_eq!(m.get_position(), &Position::new(0, 0, 0));
    m.get_position().change(1, 2);
    assert_eq!(m.get_position(), &Position::new(0, 1, 2));
    m.get_position().change(2, 1);
    assert_eq!(m.get_position(), &Position::new(0, 3, 3));
    m.get_position().set(1, 2, 3);
    m.get_position().change(2, 1);
    assert_eq!(m.get_position(), &Position::new(1, 4, 4));
}

#[test]
fn monsters_can_die() {
    let mut m = Rat::new_with_position(Position::new(0, 0, 0));
    assert_eq!(m.get_life(), 3);
    assert_eq!(m.is_dead(), false);
    m.decrease_life(1);
    assert_eq!(m.get_life(), 2);
    m.decrease_life(3);
    assert_eq!(m.get_life(), 0);
    assert_eq!(m.is_dead(), true);
}

#[test]
fn test_rat_values() {
    let m = Rat::new_with_position(Position::new(0, 0, 0));
    assert_eq!(m.monster_stats.max_life, 3);
}