use std::cmp::{max, min};

/// describes an character (PC or NPC) in the dungeon that has a position.
pub trait HasPosition {
    /// returns a mutable position
    fn get_position(&mut self) -> &mut Position;
    /// returns an immutable position
    fn get_immutable_position(&self) -> &Position;
}

#[derive(PartialEq, Debug)]
pub struct Position {
    level: usize,
    x: usize,
    y: usize,
}

impl Position {
    pub fn new(level: usize, x: usize, y: usize) -> Self {
        Self {
            level,
            x,
            y,
        }
    }
    pub fn change(&mut self, dx: i16, dy: i16) -> (usize, usize) {
        self.x = max((self.x as i16) + dx, 0) as usize;
        self.y = max((self.y as i16) + dy, 0) as usize;
        (self.x, self.y)
    }
    pub fn set(&mut self, level: usize, x: usize, y: usize) {
        self.level = level;
        self.x = x;
        self.y = y;
    }

    pub fn get_level(&self) -> usize {
        self.level
    }

    pub fn get_x(&self) -> usize {
        self.x
    }

    pub fn get_y(&self) -> usize { self.y }

    /// calculate the distance to the other position using the manhattan metric
    pub fn distance(&self, other: &Self) -> usize {
        self.x.abs_diff(other.x) + self.y.abs_diff(other.y)
    }

    /// get the direction (step size 1) towards the other position
    pub fn get_direction(&self, other: &Self) -> (i16, i16) {
        (
            max(min(other.x as i16 - self.x as i16, 1), -1),
            max(min(other.y as i16 - self.y as i16, 1), -1)
        )
    }
}

#[test]
fn test_get_direction() {
    let p1 = Position::new(0, 10, 10);
    let p2 = Position::new(0, 8, 8);
    assert_eq!(p1.get_direction(&p2), (-1, -1));
    let p2 = Position::new(0, 11, 10);
    assert_eq!(p1.get_direction(&p2), (1, 0));
}

#[test]
fn test_position_change() {
    let mut p = Position::new(0, 10, 10);

    let res = p.change(-1, 1);
    assert_eq!(res, (9, 11));
    assert_eq!(p.get_x(), 9);
    assert_eq!(p.get_y(), 11);

    let res = p.change(-9, 1);
    assert_eq!(res, (0, 12));
    assert_eq!(p.get_x(), 0);
    assert_eq!(p.get_y(), 12);

    let res = p.change(-1, -13);
    assert_eq!(res, (0, 0));
    assert_eq!(p.get_x(), 0);
    assert_eq!(p.get_y(), 0);
}

#[test]
fn test_position_set() {
    let mut p = Position::new(0, 10, 10);

    p.set(1, 2, 3);
    assert_eq!(p.get_level(), 1);
    assert_eq!(p.get_x(), 2);
    assert_eq!(p.get_y(), 3);
}